from .ConvLayer import ConvLayer
from .DenseLayer import DenseLayer
from .PoolingLayer import PoolingLayer
from .FlattenLayer import FlattenLayer
from .OutputLayer import OutputLayer
from .NetworkModel import NetworkModel
name = "inspectnn"
__version__ = "0.0.1"
__author__ = "Salvatore Barone, Filippo Ferrandino"
__credits__ = "Department of Electrical Engineering and Information Technologies, University of Naples Federico II, Via Claudio 21, Naples, Italy"
